@@ -19,9 +19,15 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
+#include <errno.h>
+#include <fcntl.h>
 #include <string.h>
 #include <stdio.h>
+#include <unistd.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 
+#include "ras-logger.h"
 #include "ras-mce-handler.h"
 #include "bitfield.h"
 
@@ -109,6 +115,7 @@ static char *mca_msg[] = {
 	[3] = "External error",
 	[4] = "FRC error",
 	[5] = "Internal parity error",
+	[6] = "SMM Handler Code Access Violation",
 };
 
 static char *tracking_msg[] = {
@@ -125,7 +132,8 @@ static const char *arstate[4] = {
 };
 
 static char *mmm_mnemonic[] = {
-	"GEN", "RD", "WR", "AC", "MS", "RES5", "RES6", "RES7"
+	"GEN", "RD",   "WR",   "AC",
+	"MS",  "RES5", "RES6", "RES7"
 };
 
 static char *mmm_desc[] = {
@@ -139,21 +147,22 @@ static char *mmm_desc[] = {
 	"Reserved 7"
 };
 
-void decode_memory_controller(struct mce_event *e, uint32_t status)
+static void decode_memory_controller(struct mce_event *e, uint32_t status)
 {
 	char channel[30];
+
 	if ((status & 0xf) == 0xf)
-		mce_snprintf(e->mc_channel, "unspecified");
+		sprintf(channel, "unspecified");
 	else
-		mce_snprintf(e->mc_channel, "%u", status & 0xf);
-	mce_snprintf(e->error_msg, "MEMORY CONTROLLER %s_CHANNEL%s_ERR\n",
-		    mmm_mnemonic[(status >> 4) & 7],
-		    channel);
-	mce_snprintf(e->error_msg, "Transaction: %s\n",
+		sprintf(channel, "%u", status & 0xf);
+
+	mce_snprintf(e->error_msg, "MEMORY CONTROLLER %s_CHANNEL%s_ERR",
+		    mmm_mnemonic[(status >> 4) & 7], channel);
+	mce_snprintf(e->error_msg, "Transaction: %s",
 		    mmm_desc[(status >> 4) & 7]);
 }
 
-static decode_termal_bank(struct mce_event *e)
+static void decode_termal_bank(struct mce_event *e)
 {
 	if (e->status & 1) {
 		mce_snprintf(e->mcgstatus_msg, "Processor %d heated above trip temperature. Throttling enabled.", e->cpu);
@@ -165,11 +174,10 @@ static decode_termal_bank(struct mce_event *e)
 
 static void decode_mcg(struct mce_event *e)
 {
-	int n, len = sizeof(e->mcgstatus_msg);
 	uint64_t mcgstatus = e->mcgstatus;
-	char *p = e->mcgstatus_msg;
 
-	mce_snprintf(e->mcgstatus_msg, "mcgstatus= %d", e->mcgstatus);
+	mce_snprintf(e->mcgstatus_msg, "mcgstatus=%lld",
+		     (long long)e->mcgstatus);
 
 	if (mcgstatus & MCG_STATUS_RIPV)
 		mce_snprintf(e->mcgstatus_msg, "RIPV");
@@ -177,6 +185,8 @@ static void decode_mcg(struct mce_event *e)
 		mce_snprintf(e->mcgstatus_msg, "EIPV");
 	if (mcgstatus & MCG_STATUS_MCIP)
 		mce_snprintf(e->mcgstatus_msg, "MCIP");
+	if (mcgstatus & MCG_STATUS_LMCE)
+		mce_snprintf(e->mcgstatus_msg, "LMCE");
 }
 
 static void bank_name(struct mce_event *e)
@@ -234,10 +244,10 @@ static void decode_mca(struct mce_event *e, uint64_t track, int *ismemerr)
 
 	if ((mca >> 2) == 3) {
 		mce_snprintf(e->mcastatus_msg,
-			     "%s Generic memory hierarchy error\n",
+			     "%s Generic memory hierarchy error",
 			     decode_attr(LL, mca & 3));
 	} else if (test_prefix(4, mca)) {
-		mce_snprintf(e->mcastatus_msg, "%s TLB %s Error\n",
+		mce_snprintf(e->mcastatus_msg, "%s TLB %s Error",
 				decode_attr(TT, (mca & TLB_TT_MASK) >> TLB_TT_SHIFT),
 				decode_attr(LL, (mca & TLB_LL_MASK) >> TLB_LL_SHIFT));
 	} else if (test_prefix(8, mca)) {
@@ -246,7 +256,7 @@ static void decode_mca(struct mce_event *e, uint64_t track, int *ismemerr)
 		char *type = decode_attr(TT, typenum);
 		char *level = decode_attr(LL, levelnum);
 		mce_snprintf(e->mcastatus_msg,
-			     "%s CACHE %s %s Error\n", type, level,
+			     "%s CACHE %s %s Error", type, level,
 			     get_RRRR_str((mca & CACHE_RRRR_MASK) >>
 					      CACHE_RRRR_SHIFT));
 #if 0
@@ -257,13 +267,13 @@ static void decode_mca(struct mce_event *e, uint64_t track, int *ismemerr)
 	} else if (test_prefix(10, mca)) {
 		if (mca == 0x400)
 			mce_snprintf(e->mcastatus_msg,
-				     "Internal Timer error\n");
+				     "Internal Timer error");
 		else
 			mce_snprintf(e->mcastatus_msg,
-				     "Internal unclassified error: %x\n",
+				     "Internal unclassified error: %x",
 				     mca);
 	} else if (test_prefix(11, mca)) {
-		mce_snprintf(e->mcastatus_msg, "BUS %s %s %s %s %s Error\n",
+		mce_snprintf(e->mcastatus_msg, "BUS %s %s %s %s %s Error",
 			     decode_attr(LL, (mca & BUS_LL_MASK) >> BUS_LL_SHIFT),
 			     decode_attr(PP, (mca & BUS_PP_MASK) >> BUS_PP_SHIFT),
 			     get_RRRR_str((mca & BUS_RRRR_MASK) >> BUS_RRRR_SHIFT),
@@ -273,7 +283,7 @@ static void decode_mca(struct mce_event *e, uint64_t track, int *ismemerr)
 		decode_memory_controller(e, mca);
 		*ismemerr = 1;
 	} else
-		mce_snprintf(e->mcastatus_msg, "Unknown Error %x\n", mca);
+		mce_snprintf(e->mcastatus_msg, "Unknown Error %x", mca);
 }
 
 static void decode_tracking(struct mce_event *e, uint64_t track)
@@ -312,7 +322,7 @@ static void decode_mci(struct mce_event *e, int *ismemerr)
 		mce_snprintf(e->mcistatus_msg, "Processor_context_corrupt");
 
 	if (e->status & (MCI_STATUS_S|MCI_STATUS_AR))
-		mce_snprintf(e->mcistatus_msg, "%s\n",
+		mce_snprintf(e->mcistatus_msg, "%s",
 			     arstate[(e->status >> 55) & 3]);
 
 	if ((e->mcgcap == 0 || (e->mcgcap & MCG_TES_P)) &&
@@ -338,6 +348,14 @@ int parse_intel_event(struct ras_events *ras, struct mce_event *e)
 	decode_mcg(e);
 	decode_mci(e, &ismemerr);
 
+	/* Check if the error is at the memory controller */
+	if (((e->status & 0xffff) >> 7) == 1) {
+		unsigned corr_err_cnt;
+
+		corr_err_cnt = EXTRACT(e->status, 38, 52);
+		mce_snprintf(e->mc_location, "n_errors=%d", corr_err_cnt);
+	}
+
 	if (test_prefix(11, (e->status & 0xffffL))) {
 		switch(mce->cputype) {
 		case CPU_P6OLD:
@@ -353,6 +371,8 @@ int parse_intel_event(struct ras_events *ras, struct mce_event *e)
 		case CPU_P4:
 			p4_decode_model(e);
 			break;
+		default:
+			break;
 		}
 	}
 	switch(mce->cputype) {
@@ -372,11 +392,85 @@ int parse_intel_event(struct ras_events *ras, struct mce_event *e)
 	case CPU_SANDY_BRIDGE_EP:
 		snb_decode_model(ras, e);
 		break;
-#if 0
 	case CPU_IVY_BRIDGE_EPEX:
 		ivb_decode_model(ras, e);
 		break;
-#endif
+	case CPU_HASWELL_EPEX:
+		hsw_decode_model(ras, e);
+	default:
+		break;
+	}
+
+	return 0;
+}
+
+/*
+ * Code to enable iMC logs
+ */
+static int domsr(int cpu, int msr, int bit)
+{
+	char fpath[32];
+	unsigned long long data;
+	int fd;
+
+	sprintf(fpath, "/dev/cpu/%d/msr", cpu);
+	fd = open(fpath, O_RDWR);
+	if (fd == -1) {
+		switch (errno) {
+		case ENOENT:
+			log(ALL, LOG_ERR,
+			    "Warning: cpu %d offline?, imc_log not set\n", cpu);
+			return -EINVAL;
+		default:
+			log(ALL, LOG_ERR,
+			    "Cannot open %s to set imc_log\n", fpath);
+			return -EINVAL;
+		}
+	}
+	if (pread(fd, &data, sizeof data, msr) != sizeof data) {
+		log(ALL, LOG_ERR,
+		    "Cannot read MSR_ERROR_CONTROL from %s\n", fpath);
+		return -EINVAL;
+	}
+	data |= bit;
+	if (pwrite(fd, &data, sizeof data, msr) != sizeof data) {
+		log(ALL, LOG_ERR,
+		    "Cannot write MSR_ERROR_CONTROL to %s\n", fpath);
+		return -EINVAL;
+	}
+	if (pread(fd, &data, sizeof data, msr) != sizeof data) {
+		log(ALL, LOG_ERR,
+		    "Cannot re-read MSR_ERROR_CONTROL from %s\n", fpath);
+		return -EINVAL;
+	}
+	if ((data & bit) == 0) {
+		log(ALL, LOG_ERR,
+		    "Failed to set imc_log on cpu %d\n", cpu);
+		return -EINVAL;
+	}
+	close(fd);
+	return 0;
+}
+
+int set_intel_imc_log(enum cputype cputype, unsigned ncpus)
+{
+	int cpu, msr, bit, rc;
+
+	switch (cputype) {
+	case CPU_SANDY_BRIDGE_EP:
+	case CPU_IVY_BRIDGE_EPEX:
+	case CPU_HASWELL_EPEX:
+		msr = 0x17f;	/* MSR_ERROR_CONTROL */
+		bit = 0x2;	/* MemError Log Enable */
+		break;
+	default:
+		return 0;
+	}
+
+	for (cpu = 0; cpu < ncpus; cpu++) {
+		rc = domsr(cpu, msr, bit);
+		if (rc)
+			return rc;
 	}
 
 	return 0;
