@@ -15,6 +15,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
+#include <ctype.h>
 #include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -25,6 +26,7 @@
 #include "ras-mce-handler.h"
 #include "ras-record.h"
 #include "ras-logger.h"
+#include "ras-report.h"
 
 /*
  * The code below were adapted from Andi Kleen/Intel/SuSe mcelog code,
@@ -45,9 +47,13 @@ static char *cputype_name[] = {
 	[CPU_SANDY_BRIDGE_EP] = "Sandy Bridge EP",	/* Fill in better name */
 	[CPU_IVY_BRIDGE] = "Ivy Bridge",		/* Fill in better name */
 	[CPU_IVY_BRIDGE_EPEX] = "Ivy Bridge EP/EX",	/* Fill in better name */
+	[CPU_HASWELL] = "Haswell",
+	[CPU_HASWELL_EPEX] = "Intel Xeon v3 (Haswell) EP/EX",
+	[CPU_BROADWELL] = "Broadwell",
+	[CPU_KNIGHTS_LANDING] = "Knights Landing",
 };
 
-enum cputype select_intel_cputype(struct ras_events *ras)
+static enum cputype select_intel_cputype(struct ras_events *ras)
 {
 	struct mce_priv *mce = ras->mce_priv;
 
@@ -71,7 +77,7 @@ enum cputype select_intel_cputype(struct ras_events *ras)
 			return CPU_NEHALEM;
 		else if (mce->model == 0x2e || mce->model == 0x2f)
 			return CPU_XEON75XX;
-		else if (mce->model == 0x2a || mce->model == 0x3a)
+		else if (mce->model == 0x2a)
 			return CPU_SANDY_BRIDGE;
 		else if (mce->model == 0x2d)
 			return CPU_SANDY_BRIDGE_EP;
@@ -79,6 +85,16 @@ enum cputype select_intel_cputype(struct ras_events *ras)
 			return CPU_IVY_BRIDGE;
 		else if (mce->model == 0x3e)
 			return CPU_IVY_BRIDGE_EPEX;
+		else if (mce->model == 0x3c || mce->model == 0x45 ||
+			 mce->model == 0x46)
+			return CPU_HASWELL;
+		else if (mce->model == 0x3f)
+			return CPU_HASWELL_EPEX;
+		else if (mce->model == 0x3d)
+			return CPU_BROADWELL;
+		else if (mce->model == 0x57)
+			return CPU_KNIGHTS_LANDING;
+
 		if (mce->model > 0x1a) {
 			log(ALL, LOG_INFO,
 			    "Family 6 Model %x CPU: only decoding architectural errors\n",
@@ -98,7 +114,7 @@ enum cputype select_intel_cputype(struct ras_events *ras)
 	return mce->family == 6 ? CPU_P6OLD : CPU_GENERIC;
 }
 
-int detect_cpu(struct ras_events *ras)
+static int detect_cpu(struct ras_events *ras)
 {
 	struct mce_priv *mce = ras->mce_priv;
 	FILE *f;
@@ -106,12 +122,12 @@ int detect_cpu(struct ras_events *ras)
 	char *line = NULL;
 	size_t linelen = 0;
 	enum {
-		VENDOR = 1,
-		FAMILY = 2,
-		MODEL = 4,
-		MHZ = 8,
-		FLAGS = 16,
-		ALL_FLAGS = 0x1f
+		CPU_VENDOR = 1,
+		CPU_FAMILY = 2,
+		CPU_MODEL = 4,
+		CPU_MHZ = 8,
+		CPU_FLAGS = 16,
+		CPU_ALL = 0x1f
 	} seen = 0;
 
 	mce->family = 0;
@@ -125,27 +141,33 @@ int detect_cpu(struct ras_events *ras)
 		return errno;
 	}
 
-	while (getdelim(&line, &linelen, '\n', f) > 0 && seen != ALL) {
-		if (sscanf(line, "mce->vendor_id : %63[^\n]", &mce->vendor) == 1)
-			seen |= VENDOR;
-		if (sscanf(line, "cpu mce->family : %d", &mce->family) == 1)
-			seen |= FAMILY;
-		if (sscanf(line, "mce->model : %d", &mce->model) == 1)
-			seen |= MODEL;
-		if (sscanf(line, "cpu MHz : %lf", &mce->mhz) == 1)
-			seen |= MHZ;
-		if (!strncmp(line, "flags", 5) && isspace(line[6])) {
+	while (seen != CPU_ALL && getdelim(&line, &linelen, '\n', f) > 0) {
+		if (sscanf(line, "vendor_id : %63[^\n]",
+		    (char *)&mce->vendor) == 1)
+			seen |= CPU_VENDOR;
+		else if (sscanf(line, "cpu family : %d", &mce->family) == 1)
+			seen |= CPU_FAMILY;
+		else if (sscanf(line, "model : %d", &mce->model) == 1)
+			seen |= CPU_MODEL;
+		else if (sscanf(line, "cpu MHz : %lf", &mce->mhz) == 1)
+			seen |= CPU_MHZ;
+		else if (!strncmp(line, "flags", 5) && isspace(line[6])) {
 			if (mce->processor_flags)
 				free(mce->processor_flags);
 			mce->processor_flags = line;
 			line = NULL;
 			linelen = 0;
-			seen |= ALL_FLAGS;
+			seen |= CPU_FLAGS;
 		}
 	}
 
-	if (!(seen == ALL)) {
-		log(ALL, LOG_INFO, "Can't parse /proc/cpuinfo\n");
+	if (seen != CPU_ALL) {
+		log(ALL, LOG_INFO, "Can't parse /proc/cpuinfo: missing%s%s%s%s%s\n",
+			(seen & CPU_VENDOR) ? "" : " [vendor_id]",
+			(seen & CPU_FAMILY) ? "" : " [cpu family]",
+			(seen & CPU_MODEL)  ? "" : " [model]",
+			(seen & CPU_MHZ)    ? "" : " [cpu MHz]",
+			(seen & CPU_FLAGS)  ? "" : " [flags]");
 		ret = EINVAL;
 		goto ret;
 	}
@@ -175,7 +197,7 @@ ret:
 	return ret;
 }
 
-int register_mce_handler(struct ras_events *ras)
+int register_mce_handler(struct ras_events *ras, unsigned ncpus)
 {
 	int rc;
 	struct mce_priv *mce;
@@ -186,15 +208,24 @@ int register_mce_handler(struct ras_events *ras)
 		return ENOMEM;
 	}
 
+	mce = ras->mce_priv;
+
 	rc = detect_cpu(ras);
 	if (rc) {
-		mce = ras->mce_priv;
 		if (mce->processor_flags)
 			free (mce->processor_flags);
 		free (ras->mce_priv);
 		ras->mce_priv = NULL;
 		return (rc);
 	}
+	switch (mce->cputype) {
+	case CPU_SANDY_BRIDGE_EP:
+	case CPU_IVY_BRIDGE_EPEX:
+	case CPU_HASWELL_EPEX:
+		set_intel_imc_log(mce->cputype, ncpus);
+	default:
+		break;
+	}
 
 	return rc;
 }
@@ -203,42 +234,10 @@ int register_mce_handler(struct ras_events *ras)
  * End of mcelog's code
  */
 
-unsigned bitfield_msg(char *buf, size_t len, char **bitarray, unsigned array_len,
-		      unsigned bit_offset, unsigned ignore_bits,
-		      uint64_t status)
-{
-	int i, n;
-	char *p = buf;
-
-	len--;
-
-	for (i = 0; i < array_len; i++) {
-		if (status & ignore_bits)
-			continue;
-		if (status & (1 <<  (i + bit_offset))) {
-			if (p != buf) {
-				n = snprintf(p, len, ", ");
-				len -= n;
-				p += n;
-			}
-			if (!bitarray[i])
-				n = snprintf(p, len, "BIT%d", i + bit_offset);
-			else
-				n = snprintf(p, len, "%s", bitarray[i]);
-			len -= n;
-			p += n;
-		}
-	}
-
-	*p = 0;
-	return p - buf;
-}
-
 static void report_mce_event(struct ras_events *ras,
 			     struct pevent_record *record,
 			     struct trace_seq *s, struct mce_event *e)
 {
-	unsigned long long val;
 	time_t now;
 	struct tm *tm;
 	struct mce_priv *mce = ras->mce_priv;
@@ -253,7 +252,7 @@ static void report_mce_event(struct ras_events *ras,
 	 */
 
 	if (ras->use_uptime)
-		now = record->ts/1000000000L + ras->uptime_diff;
+		now = record->ts/user_hz + ras->uptime_diff;
 	else
 		now = time(NULL);
 
@@ -268,19 +267,20 @@ static void report_mce_event(struct ras_events *ras,
 	else
 		trace_seq_printf(s, "bank=%x", e->bank);
 
-	trace_seq_printf(s, ", status= %d", e->status);
+	trace_seq_printf(s, ", status= %llx", (long long)e->status);
 	if (*e->error_msg)
 		trace_seq_printf(s, ", %s", e->error_msg);
 	if (*e->mcistatus_msg)
 		trace_seq_printf(s, ", mci=%s", e->mcistatus_msg);
 	if (*e->mcastatus_msg)
 		trace_seq_printf(s, ", mca=%s", e->mcastatus_msg);
-	if (*e->mc_channel)
-		trace_seq_printf(s, ", mc_channel=%s", e->mc_channel);
 
 	if (*e->user_action)
 		trace_seq_printf(s, " %s", e->user_action);
 
+	if (*e->mc_location)
+		trace_seq_printf(s, ", %s", e->mc_location);
+
 #if 0
 	/*
 	 * While the logic for decoding tsc is there at mcelog, why to
@@ -291,7 +291,7 @@ static void report_mce_event(struct ras_events *ras,
 	trace_seq_printf(s, ", walltime= %d", e->walltime);
 #endif
 
-	trace_seq_printf(s, ", CPU: %s,", cputype_name[mce->cputype]);
+	trace_seq_printf(s, ", cpu_type= %s", cputype_name[mce->cputype]);
 	trace_seq_printf(s, ", cpu= %d", e->cpu);
 	trace_seq_printf(s, ", socketid= %d", e->socketid);
 
@@ -304,28 +304,29 @@ static void report_mce_event(struct ras_events *ras,
 #endif
 
 	if (e->ip)
-		trace_seq_printf(s, ", ip= %d%s",
-				 !(e->mcgstatus & MCG_STATUS_EIPV) ? " (INEXACT)" : "",
-				 e->ip);
+		trace_seq_printf(s, ", ip= %llx%s",
+				 (long long)e->ip,
+				 !(e->mcgstatus & MCG_STATUS_EIPV) ? " (INEXACT)" : "");
 
 	if (e->cs)
-		trace_seq_printf(s, ", cs= %d", e->cs);
+		trace_seq_printf(s, ", cs= %x", e->cs);
 
 	if (e->status & MCI_STATUS_MISCV)
-		trace_seq_printf(s, ", misc= %d", e->misc);
+		trace_seq_printf(s, ", misc= %llx", (long long)e->misc);
 
 	if (e->status & MCI_STATUS_ADDRV)
-		trace_seq_printf(s, ", addr= %d", e->addr);
+		trace_seq_printf(s, ", addr= %llx", (long long)e->addr);
 
 	if (e->mcgstatus_msg)
 		trace_seq_printf(s, ", %s", e->mcgstatus_msg);
 	else
-		trace_seq_printf(s, ", mcgstatus= %d", e->mcgstatus);
+		trace_seq_printf(s, ", mcgstatus= %llx",
+				 (long long)e->mcgstatus);
 
 	if (e->mcgcap)
-		trace_seq_printf(s, ", mcgcap= %d", e->mcgcap);
+		trace_seq_printf(s, ", mcgcap= %llx", (long long)e->mcgcap);
 
-	trace_seq_printf(s, ", apicid= %d", e->apicid);
+	trace_seq_printf(s, ", apicid= %x", e->apicid);
 
 	/*
 	 * FIXME: The original mcelog userspace tool uses DMI to map from
@@ -346,7 +347,7 @@ int ras_mce_event_handler(struct trace_seq *s,
 	struct ras_events *ras = context;
 	struct mce_priv *mce = ras->mce_priv;
 	struct mce_event e;
-	int rc;
+	int rc = 0;
 
 	memset(&e, 0, sizeof(e));
 
@@ -410,5 +411,19 @@ int ras_mce_event_handler(struct trace_seq *s,
 	if (rc)
 		return rc;
 
+	if (!*e.error_msg && *e.mcastatus_msg)
+		mce_snprintf(e.error_msg, "%s", e.mcastatus_msg);
+
 	report_mce_event(ras, record, s, &e);
+
+#ifdef HAVE_SQLITE3
+	ras_store_mce_record(ras, &e);
+#endif
+
+#ifdef HAVE_ABRT_REPORT
+	/* Report event to ABRT */
+	ras_report_mce_event(ras, &e);
+#endif
+
+	return 0;
 }
